/*****************************************************************************
* amf_usb_core.c  Amfeltec USB Driver
* 		
* Copyright (C) 2013-2014, Amfeltec Corp.	
*****************************************************************************/

#ifndef __AMF_USB_CORE__
#define __AMF_USB_CORE__


#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/usb.h>

#include "amf_usb.h"
#include "amf_usb_utils.h"


//=========================================================================================================  SYNC  ============================  

int amf_usb_syncverify(amf_usb_t *ausb, int len)
{
	
	int rx_in  = ausb->rx_in;
	int rx_out = ausb->rx_out;
	int	read_index = 0;

	A_BUG(ausb == NULL);

	if (!ausb->rx_sync){

		do {
	
			A_DEBUG_TX("%s: Searching for sync from %d...len %d\n", AMF_USB_NAME, rx_out, len);
	
			if ((ausb->readbuf[rx_out] & 0xFF) != 0x7E){
				rx_out++;
				rx_out = rx_out % MAX_READ_BUF_LEN;
				len--;
				continue;
			}

			A_DEBUG_USB("%s: Got first  sync condition at %d:%d (%02X)!\n", AMF_USB_NAME, 
				rx_out, rx_in, (unsigned char)ausb->readbuf[rx_out]);


			read_index = (rx_out + 1) % MAX_READ_BUF_LEN;

			if ((ausb->readbuf[read_index] & 0xFF) != 0x5A){
				rx_out++;
				rx_out = rx_out % MAX_READ_BUF_LEN;
				len--;
				continue;			
			}

			A_DEBUG_USB("%s: Got second sync condition at %d:%d (%02X)!\n", AMF_USB_NAME, 
				read_index, rx_in,
				(unsigned char)ausb->readbuf[read_index]);

			read_index = (rx_out + 2) % MAX_READ_BUF_LEN;
		
			if ((ausb->readbuf[read_index] & 0xFF) != 0xA5){
				rx_out++;
				rx_out = rx_out % MAX_READ_BUF_LEN;
				len--;
				continue;			
			} else {
		
				A_DEBUG_USB("%s: Got third  sync condition at %d:%d (%02X)!\n", AMF_USB_NAME, 
						read_index, rx_in, (unsigned char)ausb->readbuf[read_index]);
				A_DEBUG_USB("%s: Got in sync at %d:%d! len %d\n", AMF_USB_NAME, rx_out, rx_in, len);
				ausb->rx_sync = 1;
				break;
			}	

		} while (len);

	} else {

		if ((ausb->readbuf[rx_out] & 0xFF) != 0x7E){
			A_DEBUG_USB("%s: WARNING: Got out of sync 1 at %d:%d (%02X)!\n", AMF_USB_NAME,
						rx_out, rx_in, (unsigned char)ausb->readbuf[rx_out]);
			ausb->stats.rx_sync_err_cnt++;
			ausb->rx_sync = 0;
			rx_out++;
			rx_out = rx_out % MAX_READ_BUF_LEN;
			ausb->rx_out = rx_out;
			return 0;
		}		

		read_index = (rx_out + 1) % MAX_READ_BUF_LEN;

		if ((ausb->readbuf[read_index] & 0xFF) != 0x5A) {
			A_DEBUG_USB("%s: WARNING: Got out of sync 2 at %d:%d (%02X)!\n", AMF_USB_NAME,
						rx_out, rx_in, (unsigned char)ausb->readbuf[read_index]);
			ausb->stats.rx_sync_err_cnt++;
			ausb->rx_sync = 0;
			rx_out++;
			rx_out = rx_out % MAX_READ_BUF_LEN;
			ausb->rx_out = rx_out;
			return 0;
		}		

		read_index = (rx_out + 2) % MAX_READ_BUF_LEN;

		if ((ausb->readbuf[read_index] & 0xFF) != 0xA5){
			A_DEBUG_USB("%s: WARNING: Got out of sync 3 at %d:%d (%02X)!\n", AMF_USB_NAME,
						rx_out, rx_in, (unsigned char)ausb->readbuf[read_index]);
			ausb->stats.rx_sync_err_cnt++;
			ausb->rx_sync = 0;
			rx_out++;
			rx_out = rx_out % MAX_READ_BUF_LEN;
			ausb->rx_out = rx_out;
			return 0;
		}

	}
	
	ausb->rx_out = rx_out;

	return 1;
}

int amf_usb_search_rxsync(amf_usb_t *ausb)
{
	int	retry = 0;
	do{
		if (retry++ > AMF_USBFXO_SYNC_RX_RETRIES) return -EINVAL;
		wait_just_a_bit(AMF_USBFXO_SYNC_DELAY, 0);
	} while(!ausb->rx_sync);
	return 0;
}

int amf_usb_search_txsync(amf_usb_t *ausb)
{
	int		retry = 0;
	do{
		if (retry++ > AMF_USBFXO_SYNC_TX_RETRIES) return -EINVAL;
		wait_just_a_bit(AMF_USBFXO_SYNC_DELAY, 0);
	} while(!ausb->tx_sync);
	return 0;
}
//=========================================================================================================  SYNC.



//=========================================================================================================  COMPLETION =======================

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
void amf_read_complete(struct urb *q, struct pt_regs *regs)
#else
void amf_read_complete(struct urb *q)
#endif
{
	struct amf_urb	*aurb = (struct amf_urb*)q->context;
	amf_usb_t *ausb = 		(amf_usb_t*)aurb->pvt;

	int		rx_in = 0;
	int		rx_out = 0;
	int		err, len = 0;
	int		i;

	A_BUG(ausb == NULL);

	if (q->status){
		if (!(q->status == -ENOENT ||q->status == -ECONNRESET || q->status == -ESHUTDOWN)){
			A_DEBUG_USB("%s: WARNING: RX transaction failed (%d)!\n",	AMF_USB_NAME, q->status);
		}
		if(q->status == -EPIPE) {
			/* Can't call usb_clear_halt while in_interrupt */
			ausb->stats.tx_endpointstall_cnt++;
			goto read_complete_done;
		}
		return;
	}

	ausb->rxcount++;

	if (!amf_test_bit(AMF_USB_STATUS_READY, &ausb->status)){
		A_DEBUG_USB("%s: WARNING: RX USB core is not ready (%ld:%ld)!\n", AMF_USB_NAME, 
						ausb->rxcount,(unsigned long)jiffies);               
		ausb->stats.core_notready_cnt++;
		return;
	}

	A_DEBUG_RX("%s:%d: [RX:%ld]: RX:%d:%02X %d:%02X %d:%02X (%ld)\n", 
			AMF_USB_NAME, aurb->id, ausb->rxcount, 
			q->actual_length, ((unsigned char*)q->transfer_buffer)[0], 
			ausb->next_read_ind, (unsigned char)ausb->readbuf[ausb->next_read_ind],
			ausb->next_rx_ind, (unsigned char)ausb->readbuf[ausb->next_rx_ind],
			jiffies);

	if (!q->actual_length) goto read_complete_done;

	ausb->rxbytes = ausb->rxbytes + q->actual_length;

	rx_in = ausb->rx_in;
	rx_out = ausb->rx_out;


#ifdef AMF_USB_MODE_SLAVE
	// Check master ticks
	for (i = 0; i < q->actual_length; i++) {

		if ((ausb->master_data[0] == 0) && ((ausb->readbuf_block[i] & 0xFF) == 0x7E)) {
			ausb->master_data[0] = 0x7E;
			continue;
		}

		if ((ausb->master_data[0] != 0) && (ausb->master_data[1] == 0)){

			if ((ausb->readbuf_block[i] & 0xFF) == 0x5A) {
				ausb->master_data[1] = 0x5A;
			} else {
				ausb->master_data[0] = 0;
				ausb->master_data[1] = 0;
			}
			continue;
		}

		if (ausb->master_data[1] != 0) {

			if ((ausb->readbuf_block[i] & 0xFF) == 0xA5) {
				ausb->usb_masterspan++;
				dahdi_receive(&ausb->span_master);
			} 			
			ausb->master_data[0] = 0;
			ausb->master_data[1] = 0;
		}
	}
#endif

	// Accept data
	for (i = 0; i < q->actual_length; i++) {

		ausb->readbuf[rx_in] = ausb->readbuf_block[i];

		rx_in++;
		if (rx_in == MAX_READ_BUF_LEN) { 
			rx_in = 0;	
		}
	}

	if (rx_in >= rx_out){
		len = rx_in - rx_out; 
	} else {
		len = MAX_READ_BUF_LEN - rx_out + rx_in;
	}

	if (!ausb->firmwareupdate) {
		amf_usb_syncverify(ausb, len);
	}

#ifdef AMF_USB_MODE_SLAVE

	if ((ausb->rx_sync) && (len >= MAX_USB_RX_LEN) && (!ausb->firmwareupdate) && (ausb->master_span_on == 0)){
		/* Master's tasklet */
		tasklet_schedule(&ausb->bh_task);
		ausb->usb_master++;
	}
#else

	if ((ausb->rx_sync) && (len >= MAX_USB_RX_LEN) && (!ausb->firmwareupdate)){
		/* Master's tasklet */
		tasklet_schedule(&ausb->bh_task);
		ausb->usb_master++;
	}
#endif

	ausb->rx_len = len;
	ausb->rx_in = rx_in;

read_complete_done:

	q->transfer_buffer = &ausb->readbuf_block[0];

	err = usb_submit_urb(&aurb->urb, GFP_ATOMIC);
	if ((err != 0) && (err != -ENOENT)) {
		A_DEBUG_USB("%s: usb_submit %p returns %d (%s)\n", AMF_USB_NAME, aurb, err,
					amf_test_bit(AMF_URB_STATUS_READY, &aurb->ready)?"Ready":"Busy");
	}

	return;

}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
void amf_write_complete(struct urb *q, struct pt_regs *regs)
#else
void amf_write_complete(struct urb *q)
#endif
{
	struct amf_urb	*aurb = (struct amf_urb*)q->context;
	amf_usb_t		*ausb =	(amf_usb_t*)aurb->pvt;

	A_BUG(ausb == NULL);

	ausb->txcount++;
	if (q->status){
		if (!(q->status == -ENOENT ||q->status == -ECONNRESET || q->status == -ESHUTDOWN || q->status == -EPIPE)){
			A_DEBUG("%s: WARNING: TX transaction failed (%d)!\n", AMF_USB_NAME, q->status);
		}
		if(q->status == -EPIPE) {
			/* Can't call usb_clear_halt while in_interrupt */
			ausb->stats.rx_endpointstall_cnt++;
		}
		return;
	}

	if (!amf_test_bit(AMF_USB_STATUS_READY, &ausb->status)){
		A_DEBUG_USB("%s: WARNING: TX USB core is not ready (%ld:%ld)!\n", AMF_USB_NAME, ausb->txcount,	jiffies);
		ausb->stats.core_notready_cnt++;
		return;
	}

	if ((q->actual_length < MAX_USB_TX_LEN) && (!ausb->firmwareupdate)) {
		A_DEBUG_USB("%s: WARNING: Tx length < MAX(%d:%d)\n", AMF_USB_NAME, q->actual_length, MAX_USB_TX_LEN);
	}

	/* Mark this urb as ready */
	amf_set_bit(AMF_URB_STATUS_READY, &aurb->ready);
	amf_set_bit(AMF_USB_STATUS_TX_READY, &ausb->status);

	return;

}


//=========================================================================================================  COMPLETION END



//=========================================================================================================  ENDPOINTS & URBS  ================

void amf_usb_set_transfer_urb(struct usb_interface *interface) {

	struct usb_host_interface *iface_desc; 
	struct usb_endpoint_descriptor *endpoint;
	struct amf_usb *ausb;   
	int i;
	int x;

	ausb = usb_get_intfdata(interface);

	iface_desc = interface->cur_altsetting;

	for (i = 0; i < iface_desc->desc.bNumEndpoints; i++) {
		endpoint = &iface_desc->endpoint[i].desc;	

		if (usb_endpoint_is_bulk_in(endpoint)) {

			A_DEBUG("%s: Bulk IN EP Address:%X MaxLen:%d, i=%d\n", AMF_USB_NAME, endpoint->bEndpointAddress, endpoint->wMaxPacketSize, i);	 


			for (x = 0; x < MAX_READ_URB_COUNT; x++) {
				
				usb_init_urb(&ausb->urbread[x].urb);
				usb_fill_bulk_urb (
						&ausb->urbread[x].urb,
						ausb->udev,
						usb_rcvbulkpipe (ausb->udev, endpoint->bEndpointAddress),
						&ausb->readbuf_block[64], 
						endpoint->wMaxPacketSize,                    			
						amf_read_complete, 										
						&ausb->urbread[x]);	//p									

				A_DEBUG("%s: Bulk IN EP Address:%X Len:%d\n",
						AMF_USB_NAME, 
						endpoint->bEndpointAddress,
						endpoint->wMaxPacketSize);
			}
		}				

		if (usb_endpoint_is_bulk_out(endpoint)) {

			A_DEBUG("%s: Bulk OUT EP Address:%X MaxLen:%d, i=%d\n", AMF_USB_NAME, endpoint->bEndpointAddress, endpoint->wMaxPacketSize, i);	 

			for (x = 0; x < MAX_WRITE_URB_COUNT; x++) {

				usb_init_urb(&ausb->urbwrite[x].urb);
				usb_fill_bulk_urb (
						&ausb->urbwrite[x].urb,
						ausb->udev,
						usb_sndbulkpipe (ausb->udev, endpoint->bEndpointAddress),
						&ausb->writebuf[MAX_USB_TX_LEN*x], 
						30,                          				 			
						amf_write_complete,										
						&ausb->urbwrite[x]);	//p								
																				
				ausb->urbwrite[x].next_off = MAX_USB_TX_LEN*x;
				A_DEBUG("%s: Bulk OUT EP Address:%X Len:%d\n", AMF_USB_NAME, endpoint->bEndpointAddress, 30);
			}
		}
	}
	return;
}


//=========================================================================================================   URB STATUS READY

void amf_usb_set_urb_status_ready(amf_usb_t *ausb) {
	int x;

	ausb->urbcount_read = MAX_READ_URB_COUNT;	//4
	ausb->urb_read_ind = 0;
	for (x = 0; x < ausb->urbcount_read; x++){
		ausb->urbread[x].id = x;
		ausb->urbread[x].pvt = ausb;
		amf_set_bit(AMF_URB_STATUS_READY, &ausb->urbread[x].ready);
	}

	ausb->urbcount_write = MAX_WRITE_URB_COUNT;	//4
	ausb->urb_write_ind = 0;
	for (x = 0; x < ausb->urbcount_write; x++){
		ausb->urbwrite[x].id = x;
		ausb->urbwrite[x].pvt = ausb;
		amf_set_bit(AMF_URB_STATUS_READY, &ausb->urbwrite[x].ready);
	}
	return;
}


//=========================================================================================================   Buffers ALLOC
int amf_usb_rxtx_buffers_alloc(amf_usb_t *ausb) {

	int size;
	int x;
	struct skb_buff	*skb;
	int err = -ENOMEM;

	size = MAX_WRITE_BUF_LEN;
	ausb->writebuf = kmalloc(size, GFP_ATOMIC);         
	if (ausb->writebuf == NULL){
		A_DEBUG("%s: Failed to allocate TX buffer!\n", AMF_USB_NAME);
		goto error;
	}
	memset(ausb->writebuf,0x00,size);
	amf_set_bit(AMF_USB_LSTATUS_WRITEBUF, &ausb->lstatus);

	size = MAX_READ_BUF_LEN;
	ausb->readbuf = kmalloc(size, GFP_ATOMIC);          
	if (ausb->readbuf == NULL){
		A_DEBUG("%s: Failed to allocate RX buffer!\n", AMF_USB_NAME);
		goto error;
	}
	memset(ausb->readbuf,0x00,size);
	amf_set_bit(AMF_USB_LSTATUS_READBUF, &ausb->lstatus);

	AMF_QUE_INIT(&ausb->rx_cmd_free_list);
	AMF_QUE_INIT(&ausb->rx_cmd_list);
	for(x = 0; x < AMF_USB_MAX_RX_CMD_QLEN; x++){
		skb = amf_skb_alloc(10);
		if (!skb){
			A_DEBUG("%s: Failed to allocate RX cmd buffer!\n", AMF_USB_NAME);
			goto error;
		}
		amf_skb_queue_tail(&ausb->rx_cmd_free_list, skb);
	}
	amf_set_bit(AMF_USB_LSTATUS_RXCMD, &ausb->lstatus);

	AMF_QUE_INIT(&ausb->tx_cmd_free_list);
	AMF_QUE_INIT(&ausb->tx_cmd_list);
	for(x = 0; x < AMF_USB_MAX_TX_CMD_QLEN; x++){
		skb = amf_skb_alloc(10);
		if (!skb){
			A_DEBUG("%s: Failed to allocate TX cmd buffer!\n", AMF_USB_NAME);
			goto error;
		}
		amf_skb_queue_tail(&ausb->tx_cmd_free_list, skb);
	}
	amf_set_bit(AMF_USB_LSTATUS_TXCMD, &ausb->lstatus);

	return 0;
error:
	return err;
}
//=========================================================================================================   Buffers ALLOC END


//=========================================================================================================   ENCODE / DECODE   ==========

static int amf_usb_rxcmd_decode(amf_usb_t *ausb, u8 *rx_cmd, int rx_cmd_len)
{
	struct sk_buff	*skb;
	u_int8_t		*cmd;

	A_BUG(ausb == NULL);

	switch(AMF_USB_CMD_TYPE_DECODE(rx_cmd[0])){

	case AMF_USB_CMD_TYPE_WRITE_CPU:
		A_DEBUG_TX("%s: Rx Write CPU register (%02X:%02X)!\n", AMF_USB_NAME, (unsigned char)rx_cmd[1], (unsigned char)rx_cmd[2]);
		break;			

	case AMF_USB_CMD_TYPE_READ_CPU:
		A_DEBUG_TX("%s: Rx Read CPU register (%02X:%02X)!\n", AMF_USB_NAME, (unsigned char)rx_cmd[1], (unsigned char)rx_cmd[2]);
		break;

	case AMF_USB_CMD_TYPE_WRITE_FXO:
		A_DEBUG_TX("%s: Rx Write FXO register (%02X:%02X)!\n", AMF_USB_NAME, (unsigned char)rx_cmd[1], (unsigned char)rx_cmd[2]);
		break;

	case AMF_USB_CMD_TYPE_READ_FXO:
		A_DEBUG_TX("%s: Rx Read FXO register (%d:%02X)!\n", AMF_USB_NAME, (unsigned char)rx_cmd[1],	(unsigned char)rx_cmd[2]);
		break;

	case AMF_USB_CMD_TYPE_WRITE_FXS_INDIRECT:
		A_DEBUG_TX("%s: Rx Write FXS indirect register (%d:%02X:%02X)!\n", AMF_USB_NAME, (unsigned char)rx_cmd[1], (unsigned char)rx_cmd[2], (unsigned char)rx_cmd[3]);
		break;

	case AMF_USB_CMD_TYPE_READ_FXS_INDIRECT:
		A_DEBUG_TX("%s: Rx Read FXS indirect register (%d:%02X:%02X)!\n", AMF_USB_NAME, (unsigned char)rx_cmd[1], (unsigned char)rx_cmd[2], (unsigned char)rx_cmd[3]);
		break;

	case AMF_USB_CMD_TYPE_CPU_INIT_PROSLIC:
		break;


	default:
		A_DEBUG_USB("%s: Unknown command %X : %02X!\n", AMF_USB_NAME, AMF_USB_CMD_TYPE_DECODE(rx_cmd[0]), rx_cmd[0]);
		ausb->stats.rx_cmd_unknown++;
		return -EINVAL;
	}

	if (amf_skb_queue_len(&ausb->tx_cmd_free_list)){
		skb = amf_skb_dequeue(&ausb->rx_cmd_free_list);
		cmd = amf_skb_put(skb, rx_cmd_len);
		memcpy(cmd, &rx_cmd[0], rx_cmd_len);
		amf_skb_queue_tail(&ausb->rx_cmd_list, skb);
	}

	return 0;
}




static int amf_usb_rxdata_decode (amf_usb_t *ausb, unsigned char *rxdata)
{
	int			mod_no = 0;
	u_int8_t	*rx_data[2];
	u_int8_t	rx_cmd[10];
	int		rx_cmd_len = 0;

	int reg_no = 0;
	int data_len_0 = 0;
	int i;

	u16				control_sum = 0;
	unsigned char	control_sum_low = 0;
	unsigned char	control_sum_high = 0;

	A_BUG(ausb == NULL);


	// header was already checked in sync_verify, try it again:
	if ((rxdata[0] != 0x7E) || (rxdata[1] != 0x5A) || (rxdata[2] != 0xA5)) {	
		A_DEBUG_USB("%s: WARNING: Header (%02X:%02X:%02X:%ld)!\n", AMF_USB_NAME, (unsigned int)rxdata[0], (unsigned int)rxdata[1], (unsigned int)rxdata[2], jiffies);
			ausb->stats.rx_start_fr_err_cnt++;
		return 1;
	}

	// control sum
	control_sum = 0;
	for (i = 3; i < 28; i++) {
		control_sum += rxdata[i];	
	}
	control_sum_low = control_sum & 0xFF;
	control_sum_high = (control_sum >> 8) & 0xFF;

	if ((control_sum_low != rxdata[28]) || (control_sum_high != rxdata[29])) {
		A_DEBUG_USB("%s: WARNING: Control sum (%02X:%02X: %02X:%02X: %ld)\n", AMF_USB_NAME, rxdata[28], rxdata[29], control_sum_low, control_sum_high, jiffies);
		A_DEBUG_USB("%s: Data Block: %02X:%02X:%02X:%02X:%02X %02X:%02X:%02X:%02X:%02X %02X:%02X:%02X:%02X:%02X %02X:%02X:%02X:%02X:%02X %02X:%02X:%02X:%02X:%02X %02X:%02X:%02X:%02X:%02X  rx_in:%d, rx_out:%d rx_len:%d  %ld)\n", AMF_USB_NAME, 
				rxdata[0], rxdata[1], rxdata[2], rxdata[3], rxdata[4],
				rxdata[5], rxdata[6], rxdata[7], rxdata[8], rxdata[9],
				rxdata[10], rxdata[11], rxdata[12], rxdata[13], rxdata[14],
				rxdata[15], rxdata[16], rxdata[17], rxdata[18], rxdata[19],
				rxdata[20], rxdata[21], rxdata[22], rxdata[23], rxdata[24],
				rxdata[25], rxdata[26], rxdata[27], rxdata[28], rxdata[29],
				ausb->rx_in, ausb->rx_out, ausb->rx_len,
				jiffies);

		return 1;
	} else {
		A_DEBUG_TX("%s: Control sum: OK (%02X:%02X: %02X:%02X: %ld)\n", AMF_USB_NAME, rxdata[28], rxdata[29], control_sum_low, control_sum_high, jiffies);
	}

	// MCPU Status: Frimware from CPU to PC...
	ausb->regs[0][119] = rxdata[3];			

	//	Wait for 20 seconds before MCPU errors complaining...
	if ((rxdata[3] != 0x00) && ausb->bhcount > 20000) { 
		A_DEBUG_USB("%s%d: MCPU STATUS: %02X\n", AMF_USB_NAME, ausb->order, rxdata[3]);
		ausb->stats.mcpu_status_err_cnt++;
	}

	// commands
	if (rxdata[4] != 0x7E) {

		memset(rx_cmd, 0, sizeof(rx_cmd));
		rx_cmd_len = 4;												// max for fxs indirect

		rx_cmd[0] = rxdata[4];
		rx_cmd[1] = rxdata[5];
		rx_cmd[2] = rxdata[6];
		rx_cmd[3] = rxdata[7];										// fxs indirect high

		// PROSLIC init in hardware special: we have just received proslic_init acknowlege, so it's time to reset pointers
		if (rx_cmd[0] == 0x20) {
			ausb->time_to_reset_fifo_pointers = 1;
		}

		A_DEBUG_RX("%s: Decoding RX Command 0x(%02X:%02X:%02X:%02X) %ld!\n", AMF_USB_NAME, rx_cmd[0], rx_cmd[1], rx_cmd[2], rx_cmd[3], jiffies);
		amf_usb_rxcmd_decode(ausb, rx_cmd, rx_cmd_len);

	}

	// data channel 1
	rx_data[0] = ausb->readchunk[0];
	data_len_0 = 8;

	for (i = 0; i < data_len_0; i++) {
		rx_data[0][i] = rxdata[8 + i*2];
	}

	// data channel 2
	//rx_data[1] = ausb->readchunk[1];
	//data_len_1 = (rxdata[7] >> 4) & 0xF;
	//for (i = 0; i < data_len_1; i++) {
	//	rx_data[1][i] = rxdata[9 + i*2];
	//}
	
	// data channel 2 MCPU Status:

	ausb->regs[0][120] = rxdata[9];			//dci rx read pointer
	ausb->regs[0][121] = rxdata[11];		//dci rx write pointer
	ausb->regs[0][122] = rxdata[13];		//dci rx bytes in queue

	ausb->regs[0][123] = rxdata[15];		//dci tx read pointer
	ausb->regs[0][124] = rxdata[17];		//dci tx write pointer
	ausb->regs[0][125] = rxdata[19];		//dci tx bytes in queue

	ausb->regs[0][126] = rxdata[21];		//MCPU status 2 (from hardware): 0 = Master, 1 = Slave 

	// test bit!
	if (rxdata[21] & 0x20) {
		A_DEBUG_USB("%s%d: MCPU report: INDIRECT ACCESS TIMEOUT...\n", AMF_USB_NAME, ausb->order);
		ausb->stats.slic_indirect_access_err_cnt++;
	}

	// FW2: FIFO size was 128 bytes
	// if ((ausb->regs[0][122] < 40) || (ausb->regs[0][122] > 80) || (ausb->regs[0][125] < 40) || (ausb->regs[0][125] > 80)) {
	// FW3: FIFO size is 48 bytes
	if ((ausb->regs[0][122] == 0) || (ausb->regs[0][122] >= 47) || (ausb->regs[0][125] == 0) || (ausb->regs[0][125] >= 47)) {
		ausb->stats.mcpu_queue_threshold_err_cnt++;
	}

	// CPU Registers status (for channel 0 only), 4 registers one by one round-robin every 1 ms 
	mod_no = 0;				
	reg_no = rxdata[26];
	ausb->regs[mod_no][reg_no] = rxdata[27];

	// Temperature
	ausb->temperature = rxdata[25] & 0xFF; 
	ausb->temperature |= ((rxdata[24] << 8) & 0xFF00);

	if (ausb->temperature_max < ausb->temperature) {
		ausb->temperature_max = ausb->temperature;
	}


	return 0;
}


static int amf_usb_txdata_encode (amf_usb_t *ausb, unsigned char *writebuf)
{
	unsigned char	*txdata[2];
	int				i;
	u16				control_sum = 0;

	char			test_writechunk[2][8];
	static int		test_cnt = 0;
	static char		test_char = 0x11;

#ifdef AMF_USB_MODE_SLAVE
	static unsigned long	slave_counter;	 	
#endif

	test_cnt++;
	if ((test_cnt % 2000) == 0) {
		if (test_char == 0x11) {
			test_char = 0xA5;
		} else {
			test_char = 0x11;
		}
	}
	if (((test_cnt % 10) == 0) && (test_char != 0x11)) {
		if (test_char == 0x5A) {
			test_char = 0xA5;
		} else {
			test_char = 0x5A;
		}
	}
	for (i = 0; i < 8; i++) {
		//test_writechunk[0][i] = test_char;
		//test_writechunk[1][i] = test_char;
		test_writechunk[0][i] = 0x00;
		test_writechunk[1][i] = 0x00;
	}

	A_BUG(ausb == NULL);

	memset(writebuf, 0, sizeof(writebuf));

	// header
	writebuf[0] = 0x7E;
	writebuf[1] = 0x5A;
	writebuf[2] = 0xA5;

	// Control byte: MCPU status 2: 	bit 0:	0 = Master, 1 = Slave
	//									bit 1:	0 = Idle, 	1 = FIFOs reset (on Open					
  	// set HW to Master: writebuf[3] = 0;
	writebuf[3] = ausb->master_span_on;


#ifdef AMF_USB_MODE_SLAVE
	// HW has been transfered to slave!:		1 - wait for 1000 data blocks (1 sec)
	//                               			2 - reset hardware FIFOs
	//											3 - reset err counters
	if (ausb->master_span_on == 1) {
		slave_counter++;
	} else {
		slave_counter = 0;
	}

	if ((ausb->master_span_on == 1) && (slave_counter == 1000)) {
		writebuf[3] |= 1 << 1;
		A_DEBUG_USB("%s%d: Resetting pointers: writebuf[3] = %02X\n", AMF_USB_NAME, ausb->order, writebuf[3]); 
		ausb->stats.mcpu_status_err_cnt = 0;
		ausb->stats.mcpu_queue_threshold_err_cnt = 0;
	}
	
#endif


	// DATA

	// number of data bytes(0 - 15): low: channel 0, high: channel 1 (future: dahdi-tx-buffer)
	// writebuf[7] = len & 0x0F;
	// writebuf[7] |= (len << 4) & 0xF0;
	writebuf[7] = 0x88;

	// DAHDI TX buffer:
	txdata[0] = ausb->writechunk[0];
	txdata[1] = ausb->writechunk[1];

	// DAHDI RX->TX loopback:
	//txdata[0] = ausb->readchunk[0];
	//txdata[1] = ausb->readchunk[1];

	// Test buffer:
	//txdata[0] = test_writechunk[0];
	//txdata[1] = test_writechunk[1];

	for (i = 0; i < 8; i++) {
		// channel 0:
		writebuf[8 + i*2] = txdata[0][i];
		// channel 1:
		writebuf[9 + i*2] = txdata[1][i];
	}

	// commands	
	if (amf_skb_queue_len(&ausb->tx_cmd_list)){
		u8				*cmd;
		int				cmd_len;
		struct sk_buff	*skb;
	
		skb = amf_skb_dequeue(&ausb->tx_cmd_list);
		cmd = amf_skb_data(skb);
		cmd_len = amf_skb_len(skb);

		writebuf[4] = cmd[0];
		writebuf[5] = cmd[1];
		writebuf[6] = cmd[2];
		if (cmd_len == 4) {
			writebuf[7] = cmd[3];
		}

		if (cmd_len == 2) {
			A_DEBUG_TX("%s%d: Encoding TX Command 0x(%02X:%02X) len=%d  %ld!\n", AMF_USB_NAME, ausb->order, writebuf[4], writebuf[5], cmd_len, jiffies);
		} else if (cmd_len == 3) {
			A_DEBUG_TX("%s%d: Encoding TX Command 0x(%02X:%02X:%02X) len=%d  %ld!\n", AMF_USB_NAME, ausb->order, writebuf[4], writebuf[5], writebuf[6], cmd_len, jiffies);
		} else {
			A_DEBUG_TX("%s%d: Encoding TX Command 0x(%02X:%02X:%02X:%02X) len=%d  %ld!\n", AMF_USB_NAME, ausb->order, writebuf[4], writebuf[5], writebuf[6], writebuf[7], cmd_len, jiffies);
		}

		amf_skb_init(skb, 0);
		amf_skb_queue_tail(&ausb->tx_cmd_free_list, skb);

	} else {
		writebuf[4] = AMF_USB_CTRL_IDLE_PATTERN;
	}

	// Time to reset FXS FIFO pointers in firmware, in case of:
	// - RX events enabled:	we have started DAHDI and RX events
	// - offhook:			we have received FXS off-hook status
	if (ausb->time_to_reset_fifo_pointers == 1) {
		ausb->time_to_reset_fifo_pointers = 0;
		writebuf[3] |= 1 << 1;
		A_DEBUG_USB("%s%d: Resetting pointers and MCPU stats (%02X)  %ld\n", AMF_USB_NAME, ausb->order, writebuf[3], jiffies); 
		ausb->stats.mcpu_status_err_cnt = 0;
		ausb->stats.mcpu_queue_threshold_err_cnt = 0;
	}

	// Framer status register
	writebuf[26] = 0x00;
	writebuf[27] = 0x00;

	// control sum
	control_sum = 0;
	for (i = 3; i < 28; i++) {
		control_sum += writebuf[i];
	}

	writebuf[28] = control_sum & 0xFF;
	writebuf[29] = (control_sum >> 8) & 0xFF;

	return 0;
}

//=========================================================================================================   ENCODE / DECODE.   



//=========================================================================================================    RX_TX_EVENTS   ==========

int amf_usb_rxdata_enable(amf_usb_t *ausb, int enable)
{
	A_BUG(ausb == NULL);

	if (enable){
		amf_set_bit(AMF_USB_STATUS_RX_DATA_READY, &ausb->status);
	}else{
		amf_clear_bit(AMF_USB_STATUS_RX_DATA_READY, &ausb->status);
	}
	return 0;
}


int amf_usb_rxevent_enable(amf_usb_t *ausb, int mod_no, int enable)
{
	int		event_bit;
	u8		mask;

	A_BUG(ausb == NULL);

	if (mod_no) mask = AMF_USB_CPU_BIT_CTRL_TS1_EVENT_EN;
	else mask = AMF_USB_CPU_BIT_CTRL_TS0_EVENT_EN;
	if (enable) ausb->reg_cpu_ctrl |= mask;
	else ausb->reg_cpu_ctrl &= ~mask; 
	
	__amf_usb_cpu_write(ausb, AMF_USB_CPU_REG_CTRL, ausb->reg_cpu_ctrl);

	event_bit = (mod_no) ? AMF_USB_STATUS_RX_EVENT2_READY : AMF_USB_STATUS_RX_EVENT1_READY;
	amf_set_bit(event_bit, &ausb->status);

	return 0;
}

int amf_usb_rxevent(amf_usb_t *ausb, int mod_no, int force)
{
	int		event_bit;

	A_BUG(ausb == NULL);

	event_bit = (mod_no) ? AMF_USB_STATUS_RX_EVENT2_READY : AMF_USB_STATUS_RX_EVENT1_READY;
	if(!test_bit(event_bit, (unsigned long *)&ausb->status)){
		return -EINVAL;
	}
	if (force){
		//ausb->regs[mod_no][5]  = __amf_usb_fxo_read(ausb, mod_no, 5);
		//ausb->regs[mod_no][29] = __amf_usb_fxo_read(ausb, mod_no, 29);
		//ausb->regs[mod_no][34] = __amf_usb_fxo_read(ausb, mod_no, 34);
		//ausb->regs[mod_no][4]  = __amf_usb_fxo_read(ausb, mod_no, 4);
	}

	return 0;
}

static int	amf_rx_tx_span(amf_usb_t *ausb) {

	int mod_no;

	if (ausb->isrcount % 2 == 0){
		if (ausb->mod_type[0] == AMF_USB_FXS){
			amf_proslic_check_ohttimer(ausb, 0);
			amf_proslic_check_hook(ausb, 0);
		} else {
			amf_voicedaa_check_hook(ausb, 0);
		}
	}

	if (!(ausb->isrcount & 0xF0)){
		if (ausb->mod_type[0] == AMF_USB_FXS){
			amf_proslic_recheck_sanity(ausb, 0);
		} else {
			amf_voicedaa_recheck_sanity(ausb, 0);
		}
	}

	if (amf_test_bit(AMF_USB_LSTATUS_DAHDI, &ausb->lstatus)){

		for (mod_no = 0; mod_no < ausb->mod_cnt; mod_no++) {
			dahdi_ec_chunk(&ausb->chans[mod_no], ausb->chans[mod_no].readchunk, ausb->lasttx[mod_no]);
		}

		dahdi_receive(&ausb->span);
		dahdi_transmit(&ausb->span);

		for (mod_no = 0; mod_no < ausb->mod_cnt; mod_no++) {
			memcpy(ausb->lasttx[mod_no], ausb->chans[mod_no].writechunk, DAHDI_CHUNKSIZE);
		}

	}

	return 0;
}


//=========================================================================================================    RX_TX_EVENTS END



//=========================================================================================================    ISR   ==========

void amf_usb_isr(void *arg)
{
	amf_usb_t		*ausb = (amf_usb_t*)arg;
	int i;

	ausb->isrcount++;

	for (i = 0; i < ausb->mod_cnt; i++) {
		
		amf_usb_rxevent(ausb, i, 0);

	}

	amf_rx_tx_span(ausb);

	return;
}



//=========================================================================================================  TASKLET  ==========


int amf_usb_txdata_prepare(amf_usb_t *ausb, char *writebuf, int len)
{
	int		tx_in;
	int		tx_out; 
	int		urb_write_ind;
	int		err;

	A_BUG(ausb == NULL);
	
	tx_in = ausb->tx_in;
	tx_out = ausb->tx_out;
	
	urb_write_ind = (ausb->urb_write_ind + 1) % MAX_WRITE_URB_COUNT;

	memcpy(&ausb->writebuf[tx_in], writebuf, len); 
	tx_in = (tx_in + len) % MAX_WRITE_BUF_LEN;  

	if (!amf_test_bit(AMF_URB_STATUS_READY, &ausb->urbwrite[urb_write_ind].ready)){
		/* Use previous urb for now and increment errors */
		A_DEBUG_TX("%s: [BH:%ld:%ld:%ld]: TX is not ready (%ld)!\n",
				AMF_USB_NAME, 
				ausb->bhcount,
				ausb->rxcount,
				ausb->txcount,
				(unsigned long)jiffies);
		ausb->stats.tx_overrun_cnt++;
		return -EINVAL;
	}

	/* Update tx urb */
	ausb->urbwrite[urb_write_ind].urb.transfer_buffer = &ausb->writebuf[tx_out];
	ausb->urbwrite[urb_write_ind].urb.transfer_buffer_length = len;

	tx_out = (tx_out + len) % MAX_WRITE_BUF_LEN;  
	amf_clear_bit(AMF_USB_STATUS_TX_READY, &ausb->status);

	err = usb_submit_urb(&ausb->urbwrite[urb_write_ind].urb, GFP_ATOMIC); 
	if (err != 0) {
		if (err != -ENOENT) {
			A_DEBUG("%s: Failed to program transmitter (%ld)!\n", AMF_USB_NAME, jiffies);
		}
		ausb->stats.tx_notready_cnt++;
		return  -EINVAL;
	}

	ausb->urb_write_ind = urb_write_ind;
	ausb->tx_in = tx_in;
	ausb->tx_out = tx_out;

	return 0;

}


static int amf_usb_rxdata_get(amf_usb_t *ausb, unsigned char *rxdata, int max_len)
{
	int		rx_len;
	int 	i;
	int		rx_out = 0;
	int		rx_in = 0;
	
	
	A_BUG(ausb == NULL);

	rx_in = ausb->rx_in;
	rx_out = ausb->rx_out;

	if (rx_in > rx_out){
		rx_len = rx_in - rx_out; 
	}else if (rx_in < rx_out){					
		rx_len = MAX_READ_BUF_LEN - rx_out + rx_in;
	}else{
		return 0;
	}

	if (rx_len > max_len) rx_len = max_len;                         
	if (rx_len > MAX_USB_RX_LEN) rx_len = MAX_USB_RX_LEN;

	for (i = 0; i < rx_len; i++) {

		rxdata[i] = ausb->readbuf[rx_out];
		rx_out++;
		if (rx_out == MAX_READ_BUF_LEN) {
			rx_out = 0;
		}		
	}

	ausb->rx_out = rx_out;
	
	return rx_len;

}

static int amf_usb_bh_rx (amf_usb_t *ausb)
{
	int			rx_len;
	u_int8_t	readbuf[MAX_USB_RX_LEN];
	int 		datablocks_num;

	A_BUG(ausb == NULL);

	memset(readbuf, 0, sizeof(readbuf));

	rx_len = amf_usb_rxdata_get(ausb, &readbuf[0], MAX_USB_RX_LEN);
	if (!rx_len){
		A_DEBUG_USB("%s: RX: No data available! (%d:%d:%d:%d)\n", AMF_USB_NAME, rx_len,
				ausb->rx_out,
				ausb->rx_in,
				MAX_READ_BUF_LEN);
		ausb->stats.rx_underrun_cnt++;
		return -EINVAL;
	}

	if (rx_len < MAX_USB_RX_LEN){
		A_DEBUG_USB("%s: RX: Not enough data! (%d:%d:%d:%d)\n", AMF_USB_NAME, rx_len,
				ausb->rx_out,
				ausb->rx_in,
				MAX_READ_BUF_LEN);
		ausb->stats.rx_underrun_cnt++;
		return 0;
	}

	datablocks_num = rx_len % MAX_USB_RX_LEN;

	if (datablocks_num > 5){
		A_DEBUG_USB("%s: RX: Too slow: more then 5 blocks in a buffer! (%d:%d:%d:%d)\n", AMF_USB_NAME, rx_len,
				ausb->rx_out,
				ausb->rx_in,
				MAX_READ_BUF_LEN);
		ausb->stats.rx_overrun_cnt++;
		return 0;
	}

	amf_usb_rxdata_decode(ausb, readbuf);

	if (amf_test_bit(AMF_USB_STATUS_RX_DATA_READY, &ausb->status) && ausb->isr_func){
		ausb->isr_func(ausb->isr_arg);
	}

	return 0;
}


static int amf_usb_bh_tx (amf_usb_t *ausb)
{	
	unsigned char	writebuf[MAX_USB_TX_LEN];

	A_BUG(ausb == NULL);
	
	memset(writebuf, 0, sizeof(writebuf));
	amf_usb_txdata_encode(ausb, writebuf);

 	if (amf_usb_txdata_prepare(ausb, writebuf, MAX_USB_TX_LEN)){
		return -EINVAL;
	}

	return 0;
}


void amf_usb_bh (unsigned long data)
{
	amf_usb_t		*ausb = (amf_usb_t*)data;
	
	A_BUG(ausb == NULL);

	if (!amf_test_bit(AMF_USB_STATUS_READY, &ausb->status)){
		goto usb_bh_exit;
	}
	
	if (amf_test_and_set_bit(AMF_USB_STATUS_BH, &ausb->status)){
		A_DEBUG("%s: [BH:%ld]: Re-entry in USB BH!\n", AMF_USB_NAME, jiffies);
		goto usb_bh_done;
	}
	ausb->bhcount++;


	if ((ausb->mod_type[0] == AMF_USB_FXO) && (ausb->failover_enabled)) {

		if ((ausb->bhcount % 5000) == 0) {
			A_DEBUG_TX("%s%d: WATCHDOG: %ld\n", AMF_USB_NAME, ausb->order, jiffies);
			__amf_usb_cpu_write(ausb, AMF_USB_CPU_REG_WATCHDOG, AMF_USB_CPU_WATCHDOG_RESTART);	
		}	
	}

	// If we are called by Master, do receive and transmit!
	// If we are called by Slave: data will be checked in get!

	/* receive path */
	amf_usb_bh_rx(ausb);

	/* transmit path */
	amf_usb_bh_tx(ausb);

	amf_clear_bit(AMF_USB_STATUS_BH, &ausb->status);

usb_bh_done:

usb_bh_exit:
	return;
}

//=========================================================================================================  TASKLET (end)



//=========================================================================================================  RAW TX/RX  =========

int amf_usb_txdata_raw(amf_usb_t *ausb, unsigned char *data, int max_len)
{
	int		err = 0;
	
	// raw read/write is still in 32-byte frames
	if (max_len > 32){
		max_len = 32;				
	}

	err = amf_usb_txdata_prepare(ausb, data, max_len);

	return (!err) ? max_len : 0;
}

int amf_usb_txdata_raw_ready(amf_usb_t *ausb)
{
	if (!amf_test_bit(AMF_USB_STATUS_TX_READY, &ausb->status)){
		return -EBUSY;
	}
	return 0;
}

int amf_usb_rxdata_raw(amf_usb_t *ausb, unsigned char *data, int max_len)
{
	int		rxlen = 0;

	// raw read/write is still in 32-byte frames
	if (max_len > 32){
		max_len = 32;
	}

	rxlen = amf_usb_rxdata_get(ausb, data, max_len);
	return rxlen;
}


int amf_usb_rxurb_reset(amf_usb_t *ausb)
{
	int		x = 0;
	int		err = 0;

	ausb->rx_in	= 0;
	ausb->rx_out = 0;
	for (x = 0; x < ausb->urbcount_read; x++){

		ausb->urbread[x].urb.transfer_buffer = &ausb->readbuf_block[0];

		err = usb_submit_urb(&ausb->urbread[x].urb, GFP_KERNEL); 
		if ((err != 0) && (err != -ENOENT)) {
			A_DEBUG("%s: Failed to start Raw RX transfer: %d\n", AMF_USB_NAME, x);
			return  -EINVAL;
		}

	}
	return 0;
}

int amf_usb_txurb_reset(amf_usb_t *ausb)
{
	ausb->tx_in = 0;
	ausb->tx_out = 0;
	return 0;
}


#endif  //__AMF_USB_CORE__


